(*
 *                     The OCaml-gtk interface
 *
 * Copyright (c) 1997, 1998   David Monniaux & Pascal Cuoq
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 * [see the file COPYING.LIB]
 *)

(* $Id: glib.ml,v 1.3 1998/11/20 10:40:17 monniaux Exp $ *)

(* WARNING:
These lists are NOT for storing ML boxed values.
The pointers in those lists won't be updated during garbage collection.
*)
module GList =
struct
  type 'a t
  external empty: unit->'a t = "mlg_zero";;
  external alloc: unit->'a t = "g_list_alloc";;
  external free: 'a t->unit = "mlg_list_free";;
  external free_1: 'a t->unit = "mlg_list_free_1";;
  external append: 'a t->'a->'a t = "g_list_append";;
  external prepend: 'a t->'a->'a t = "g_list_prepend";;
  external insert: 'a t->'a->int->'a t = "mlg_list_insert";;
  external remove: 'a t->'a->'a t = "g_list_remove";;
  external remove_link: 'a t->'a t->'a t="g_list_remove_link";;
  external reverse: 'a t->'a t="g_list_reverse";;
  external nth: 'a t->int->'a t="mlg_list_nth";;
  external find: 'a t->'a->'a t="g_list_find";;
  external last: 'a t->'a ="g_list_last";;
  external first: 'a t->'a ="g_list_first";;
  external length: 'a t->int="mlg_list_length";;
  external foreach: 'a t->('a->unit)->unit="mlg_list_foreach";;

  let rec from_ml: 'a list->'a t = function
      [] -> empty()
    | h::t -> prepend (from_ml t) h;;
  let ml_from (l : 'a t) =
    let output=ref [] in
      foreach l (fun x->output:=x :: !output);
      List.rev(! output);;
end;;

module GSList =
struct
  type 'a t
  external empty: unit->'a t = "mlg_zero";;
  external alloc: unit->'a t = "g_slist_alloc";;
  external free: 'a t->unit = "mlg_slist_free";;
  external free_1: 'a t->unit = "mlg_slist_free_1";;
  external append: 'a t->'a->'a t = "g_slist_append";;
  external prepend: 'a t->'a->'a t = "g_slist_prepend";;
  external insert: 'a t->'a->int->'a t = "mlg_slist_insert";;
  external remove: 'a t->'a->'a t = "g_slist_remove";;
  external remove_link: 'a t->'a t->'a t="g_slist_remove_link";;
  external reverse: 'a t->'a t="g_slist_reverse";;
  external nth: 'a t->int->'a t="mlg_slist_nth";;
  external find: 'a t->'a->'a t="g_slist_find";;
  external last: 'a t->'a ="g_slist_last";;
  (*external first: 'a t->'a ="g_slist_first";;*)
  external length: 'a t->int="mlg_slist_length";;
  external foreach: 'a t->('a->unit)->unit="mlg_slist_foreach";;

  let rec from_ml: 'a list->'a t = function
      [] -> empty()
    | h::t -> prepend (from_ml t) h;;
  let ml_from (l : 'a t) =
    let output=ref [] in
      foreach l (fun x->output:=x :: !output);
      List.rev(! output);;
end;;

module Pointer = struct
  type t
  external int_poke : t->int->unit = "mlgtk_int_poke"
  external int_peek : t->int = "mlgtk_int_peek"
end;;
