/*
 *                     The OCaml-gtk interface
 *
 * Copyright (c) 1997, 1998   David Monniaux & Pascal Cuoq
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 * [see the file COPYING.LIB]
 */

/* $Id: mlgdk_stub.c,v 1.11 1998/11/20 10:40:19 monniaux Exp $ */

#include <caml/memory.h>
#include <caml/mlvalues.h>
#include <gdk/gdk.h>
#include "config.h"

/* Events */

#define mlgdk_expose_event_extract(coord) \
value mlgdk_expose_event_extract_##coord (value ptr) \
{ \
  return Val_int(((GdkEventExpose*) ptr)->area.##coord); \
}
mlgdk_expose_event_extract(x)
mlgdk_expose_event_extract(y)
mlgdk_expose_event_extract(width)
mlgdk_expose_event_extract(height)

value mlgdk_key_event_extract_string(value ptr)
{
  return copy_string(((GdkEventKey*) ptr)->string);
}


value mlgdk_key_event_extract_keyval(value ptr)
{
  return Val_int(((GdkEventKey*) ptr)->keyval);
}


value mlgdk_key_event_extract_state(value ptr)
{
  return Val_int(((GdkEventKey*) ptr)->state);
}

#define mlgdk_refoption_ret(id)\
  if (ml_##id != 1) Field(Field(ml_##id, 0),0)=(value) id

#define mlgdk_refoption_prm(id) ((ml_##id == 1) ? NULL : &id)
#define mlgdk_option_prm(id) ((ml_##id == 1) ? NULL : (void*)Field(ml_##id, 0))

/* Color */
#define GdkColor_ml(x) (*((GdkColor*) (&Field(x, 0))))
static value exn_not_available;;

#define mlgdk_color_ml_alloc() \
  alloc(Abstract_tag, (sizeof(GdkColor)+sizeof(void*)-1)/sizeof(void*))

value mlgdk_color_alloc(value map, value rgb)
{
  GdkColor gdkcolor;
  value retval;

  gdkcolor.red = Field(rgb, 0);
  gdkcolor.green = Field(rgb, 1);
  gdkcolor.blue = Field(rgb, 2);

  if (! gdk_color_alloc((GdkColormap*)map, &gdkcolor))
    raise_constant(exn_not_available);
   
  retval = mlgdk_color_ml_alloc();
  GdkColor_ml(retval) = gdkcolor;
  return retval;
}

/* Fonts */
value mlgdk_font_load(value font_name)
{
  return (value) gdk_font_load(String_val(font_name));
}

/* GC */
value mlgdk_gc_destroy(value gc)
{
  gdk_gc_destroy((GdkGC*) gc);
  return Val_unit;
}

value mlgdk_gc_set_clip_rectangle(value gc,
  value x, value y, value w, value h)
{
  GdkRectangle r = {Int_val(x), Int_val(y), Int_val(w), Int_val(h)};
  gdk_gc_set_clip_rectangle((GdkGC*) gc, &r);
  return Val_unit;
}

value mlgdk_gc_set_foreground(value gc, value color)
{
  gdk_gc_set_foreground((GdkGC*)gc, &GdkColor_ml(color));
  return Val_unit;
}

value mlgdk_gc_set_background(value gc, value color)
{
  gdk_gc_set_background((GdkGC*)gc, &GdkColor_ml(color));
  return Val_unit;
}

value mlgdk_gc_set_font(value gc, value font)
{
  gdk_gc_set_font((GdkGC*)gc, (GdkFont*) font);
  return Val_unit;
}

value mlgdk_gc_set_function(value gc, value function)
{
  gdk_gc_set_function((GdkGC*)gc, (GdkFunction) (Int_val(function)));
  return Val_unit;
}

value mlgdk_gc_set_fill(value gc, value fill)
{
  gdk_gc_set_fill((GdkGC*) gc, Int_val(fill));
  return Val_unit;
}

value mlgdk_gc_set_tile(value gc, value pixmap)
{
  gdk_gc_set_tile((GdkGC*) gc, (GdkPixmap*) pixmap);
  return Val_unit;
}

value mlgdk_gc_set_stipple(value gc, value fill)
{
  gdk_gc_set_fill((GdkGC*)gc, (GdkFill) (Int_val(fill)));
  return Val_unit;
}

value mlgdk_gc_set_line_attributes(value gc, value line_width,
  value line_style, value cap_style, value join_style)
{
  gdk_gc_set_line_attributes((GdkGC*) gc,
    Int_val(line_width), Int_val(line_style),
    Int_val(cap_style), Int_val(join_style));
  return Val_unit;
}

/* Pixmap */
value mlgdk_pixmap_new(value window, value width, value height, value depth)
{
  return (value) gdk_pixmap_new((GdkWindow*) window,
    Int_val(width), Int_val(height), Int_val(depth));
}

value mlgdk_pixmap_create_from_xpm(value window, value ml_mask,
  value ml_transparent_color, value filename)
{
  value r;
  GdkBitmap *mask;
  r = (value) gdk_pixmap_create_from_xpm((GdkWindow*) window,
    mlgdk_refoption_prm(mask),
    mlgdk_option_prm(transparent_color),
    String_val(filename));
  mlgdk_refoption_ret(mask);
  return r;
}

/* Drawing */
value mlgdk_draw_point(value drawable, value gc, value x, value y)
{
  gdk_draw_point((GdkDrawable*)drawable, (GdkGC*)gc, Int_val(x), Int_val(y));
  return Val_unit;
}

value mlgdk_draw_line_native(value drawable, value gc,
  value x1, value y1, value x2, value y2)
{
  gdk_draw_line((GdkDrawable*)drawable, (GdkGC*)gc,
   Int_val(x1), Int_val(y1), Int_val(x2), Int_val(y2));
  return Val_unit;
}

value mlgdk_draw_line(value *argv, int argc)
{
  return mlgdk_draw_line_native(argv[0], argv[1],
    argv[2], argv[3], argv[4], argv[5]);
}

value mlgdk_draw_rectangle_native(value drawable, value gc, value filled,
  value x, value y, value w, value h)
{
  gdk_draw_rectangle((GdkDrawable*)drawable, (GdkGC*)gc, Int_val(filled),
   Int_val(x), Int_val(y), Int_val(w), Int_val(h));
  return Val_unit;
}

value mlgdk_draw_rectangle(value *argv, int argc)
{
  return mlgdk_draw_rectangle_native(argv[0], argv[1],
    argv[2], argv[3], argv[4], argv[5], argv[6]);
}

value mlgdk_draw_arc_native(value drawable, value gc, value filled,
  value x, value y, value w, value h,
  value angle1, value angle2)
{
  gdk_draw_arc((GdkDrawable*)drawable, (GdkGC*)gc, Int_val(filled),
   Int_val(x), Int_val(y), Int_val(w), Int_val(h),
   Int_val(angle1), Int_val(angle2));
  return Val_unit;
}

value mlgdk_draw_arc(value *argv, int argc)
{
  return mlgdk_draw_arc_native(argv[0], argv[1],
    argv[2], argv[3], argv[4], argv[5], argv[6],
    argv[7], argv[8]);
}

value mlgdk_draw_polygon(value drawable, value gc, value filled,
  value xyarray)
{
  int n = Wosize_val(xyarray);
  value xy;
  BEGIN_LOCAL_ARRAY(GdkPoint, xyarray_g, n);
  int i;
  for (i=0; i<n; i++)
    {
      value xy = Field(xyarray, i);
      xyarray_g[i].x = Field(xy, 0);
      xyarray_g[i].y = Field(xy, 1);
    }
  gdk_draw_polygon((GdkDrawable*)drawable, (GdkGC*)gc,
    Int_val(filled), xyarray_g, n);
  END_LOCAL_ARRAY(GdkPoint, xyarray_g, n);
  return Val_unit;
}

value mlgdk_draw_string_native(value drawable, value font, value gc,
  value x, value y, value string)
{
  gdk_draw_string((GdkDrawable*)drawable, (GdkFont*)font, (GdkGC*)gc,
    Int_val(x), Int_val(y), String_val(string));
  return Val_unit;
}

value mlgdk_draw_string(value *argv, int argc)
{
  return mlgdk_draw_string_native(argv[0], argv[1],
    argv[2], argv[3], argv[4], argv[5]);
}

value mlgdk_draw_pixmap_native(value drawable, value gc, value src,
  value xsrc, value ysrc, value xdest, value ydest, value width, value height)
{
  gdk_draw_pixmap((GdkDrawable*)drawable, (GdkGC*) gc, (GdkPixmap*) src,
    Int_val(xsrc), Int_val(ysrc), Int_val(xdest), Int_val(ydest),
    Int_val(width), Int_val(height));
  return Val_unit;
}

value mlgdk_draw_pixmap(value *argv, int argc)
{
  return mlgdk_draw_pixmap_native(argv[0], argv[1],
    argv[2], argv[3], argv[4], argv[5], argv[6], argv[7], argv[8]);
}

value mlgdk_draw_points(value drawable, value gc, value xyarray)
{
  int n = Wosize_val(xyarray);
  value xy;
  BEGIN_LOCAL_ARRAY(GdkPoint, xyarray_g, n);
  int i;
  for (i=0; i<n; i++)
    {
      value xy = Field(xyarray, i);
      xyarray_g[i].x = Field(xy, 0);
      xyarray_g[i].y = Field(xy, 1);
    }
  gdk_draw_points((GdkDrawable*)drawable, (GdkGC*)gc, xyarray_g, n);
  END_LOCAL_ARRAY(GdkPoint, xyarray_g, n);
  return Val_unit;
}

value mlgdk_draw_segments(value drawable, value gc, value xyarray)
{
  int n = Wosize_val(xyarray);
  value xy;
  BEGIN_LOCAL_ARRAY(GdkSegment, xyarray_g, n);
  int i;
  for (i=0; i<n; i++)
    {
      value xy = Field(xyarray, i);
      xyarray_g[i].x1 = Field(xy, 0);
      xyarray_g[i].y1 = Field(xy, 1);
      xyarray_g[i].x2 = Field(xy, 2);
      xyarray_g[i].y2 = Field(xy, 3);
    }
  gdk_draw_segments((GdkDrawable*)drawable, (GdkGC*)gc, xyarray_g, n);
  END_LOCAL_ARRAY(GdkSegment, xyarray_g, n);
  return Val_unit;
}

value mlgdk_draw_lines(value drawable, value gc, value xyarray)
{
  int n = Wosize_val(xyarray);
  value xy;
  BEGIN_LOCAL_ARRAY(GdkPoint, xyarray_g, n);
  int i;
  for (i=0; i<n; i++)
    {
      value xy = Field(xyarray, i);
      xyarray_g[i].x = Field(xy, 0);
      xyarray_g[i].y = Field(xy, 1);
    }
  gdk_draw_lines((GdkDrawable*)drawable, (GdkGC*)gc, xyarray_g, n);
  END_LOCAL_ARRAY(GdkPoint, xyarray_g, n);
  return Val_unit;
}

/* Initialization */
void mlgdk_init()
{
  exn_not_available = caml_named_value("color not available");
}
