/*
 *                     The OCaml-gtk interface
 *
 * Copyright (c) 1997, 1998   David Monniaux & Pascal Cuoq
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 * [see the file COPYING.LIB]
 */

/* $Id: mlgtk_stub.c,v 1.26 1998/12/21 14:41:36 monniaux Exp $ */

#include <caml/mlvalues.h>
#include <caml/memory.h>
#include <gtk/gtk.h>

#define Double_array_length(v) (Wosize_val(v)/(sizeof(double)/sizeof(value)))

/* ###### 3. Objects */
/* ### 3.1 Type utility functions */
/* ### 3.2 Object functions */

#define GtkCurve_ml(object) (GTK_CURVE(object))
#define GtkCombo_ml(object) (GTK_COMBO(object))
#define GtkItem_ml(object) (GTK_ITEM(object))
#define GtkAlignment_ml(object) (GTK_ALIGNMENT(object))
#define GtkAspectFrame_ml(object) (GTK_ASPECT_FRAME(object))
#define GtkArrow_ml(object) (GTK_ARROW(object))
#define GtkButtonBox_ml(object) (GTK_BUTTON_BOX(object))
#define GtkObject_ml(object) (GTK_OBJECT(object))
#define GtkWidget_ml(object) (GTK_WIDGET(object))
#define GtkBox_ml(object) (GTK_BOX(object))
#define GtkCList_ml(object) (GTK_CLIST(object))
#define GtkList_ml(object) (GTK_LIST(object))
#define GtkMenu_ml(object) (GTK_MENU(object))
#define GtkMenuItem_ml(object) (GTK_MENU_ITEM(object))
#define GtkMenuBar_ml(object) (GTK_MENU_BAR(object))
#define GtkContainer_ml(object) (GTK_CONTAINER(object))
#define GtkPaned_ml(object) (GTK_PANED(object))
#define GtkTooltips_ml(object) (GTK_TOOLTIPS(object))
#define GtkWindow_ml(object) (GTK_WINDOW(object))
#define GtkButton_ml(object) (GTK_BUTTON(object))
#define GtkToggleButton_ml(object) (GTK_TOGGLE_BUTTON(object))
#define GtkProgressBar_ml(object) (GTK_PROGRESS_BAR(object))
#define GtkText_ml(object) (GTK_TEXT(object))
#define GtkDrawingArea_ml(object) (GTK_DRAWING_AREA(object))
#define GtkColorSelection_ml(object) (GTK_COLOR_SELECTION(object))
#define GtkNotebook_ml(object) (GTK_NOTEBOOK(object))
#define GtkEditable_ml(object) (GTK_EDITABLE(object))
#define GtkEntry_ml(object) (GTK_ENTRY(object))
#define GtkFrame_ml(object) (GTK_FRAME(object))
#define GtkCheckMenuItem_ml(object) (GTK_CHECK_MENU_ITEM(object))
#define ml_GtkObject(object) ((value) object)

#define check_GtkObject_ml(object, func_id)
/* if (GtkObject_ml(object) == NULL) \
    invalid_argument(func_id) */

#define optionptr_ml(x) (Is_long(x) ? NULL : (void*) Field(x, 0))

value mlgtk_window_of(value widget)
{
  return (value) (((GtkWidget*) widget)->window);
}

value mlgtk_style_of(value widget)
{
  value r=alloc_tuple(1);
  GtkStyle *style=((GtkWidget*)widget)->style;
  int i=0;
  Field(r, i++) = (value) (style->font);
  Field(r, i++) = (value) (style->black_gc);
  Field(r, i++) = (value) (style->colormap);
  return r;
}

value mlgtk_int_poke(value ptr, value val)
{
  *((gint*) ptr) = Int_val(val);
  return Val_unit;
}

value mlgtk_int_peek(value ptr)
{
  return Val_int(*((gint*) ptr));
}

/* ###### 4. Signals */
void mlgtk_callback_exec(GtkObject *object, value *closurep,
  int n_args, GtkArg *args)
{
  int i;
  value retval, car, cdr;

  /*fprintf(stderr, "callback with %d args\n", n_args);*/
  Begin_roots1(retval);
  retval=Val_int(0);
  for (i=n_args-1; i>=0; i--)
    {
      fprintf(stderr, "arg #%d type=%d\n", i,
        GTK_FUNDAMENTAL_TYPE(args[i].type));
      switch (GTK_FUNDAMENTAL_TYPE(args[i].type))
	{
	case GTK_TYPE_NONE:
	  car = Val_int(0);
	  break;

	case GTK_TYPE_INVALID:
	  car = Val_int(2);
	  break;

	case GTK_TYPE_INT:
	case GTK_TYPE_UINT:
	  car = alloc(1, 2); /* Int */
	  Field(car, 0) = Val_int(args[i].d.int_data);
	  break;

	case GTK_TYPE_LONG:
	case GTK_TYPE_ULONG:
	  car = alloc(1, 2); /* Int */
	  Field(car, 0) = Val_int(args[i].d.long_data);
	  break;

	case GTK_TYPE_BOOL:
	  car = alloc(1, 0); /* Bool */
	  Field(car, 0) = Val_bool(args[i].d.int_data);
	  break;

	case GTK_TYPE_CHAR:
	  car = alloc(1, 1); /* Char */
	  Field(car, 0) = Val_int(args[i].d.char_data);
	  break;

	case GTK_TYPE_FLOAT:
	  car = alloc(1, 3); /* Float */
	  Field(car, 0) = copy_double(args[i].d.float_data);
	  break;

	case GTK_TYPE_STRING:
	  car = alloc(1, 4); /* Float */
	  Field(car, 0) = copy_string(args[i].d.string_data);
	  break;

	case GTK_TYPE_POINTER:
	case GTK_TYPE_BOXED:
	  car = alloc(1, 5); /* Pointer */
	  Field(car, 0) = (value) args[i].d.pointer_data;
	  break;

	default: car = Val_int(1); /* Other */
	}
      cdr=retval;
      retval = alloc(2, 0);
      Field(retval, 0)=car;
      Field(retval, 1)=cdr;
    }
  fflush(stderr);

  retval = callback(*closurep, retval);
  End_roots();

  if (Is_long(retval))
    {
      switch (Int_val(retval))
	{
	case 0: /* unit */
	  switch (GTK_FUNDAMENTAL_TYPE(args[n_args].type))
	    {
	    case GTK_TYPE_NONE:
	      return;
	    }
	}
    }
  else
    {
      switch (Tag_val(retval))
	{
	case 0: /* bool */
	  switch (GTK_FUNDAMENTAL_TYPE(args[n_args].type))
	    {
	    case GTK_TYPE_BOOL:
	      *((gint*) (args[n_args].d.pointer_data)) = Int_val(Field(retval, 0));
	      return;
	    }

	case 1: /* char */
	  switch (GTK_FUNDAMENTAL_TYPE(args[n_args].type))
	    {
	    case GTK_TYPE_CHAR:
	      *((gchar*) (args[n_args].d.pointer_data)) = Int_val(Field(retval, 0));
	      return;
	    }

	case 2: /* int */
	  switch (GTK_FUNDAMENTAL_TYPE(args[n_args].type))
	    {
	    case GTK_TYPE_ENUM:
	    case GTK_TYPE_INT:
	    case GTK_TYPE_UINT:
	      *((gint*) (args[n_args].d.pointer_data)) = Int_val(Field(retval, 0));
	      return;

	    case GTK_TYPE_LONG:
	    case GTK_TYPE_ULONG:
	      *((glong*) (args[n_args].d.pointer_data)) = Int_val(Field(retval, 0));
	      return;
	    }

	case 3: /* float */
	  switch (GTK_FUNDAMENTAL_TYPE(args[n_args].type))
	    {
	    case GTK_TYPE_FLOAT:
	      *((gfloat*) (args[n_args].d.pointer_data)) = Double_val(Field(retval, 0));
	      return;
	    }

	case 4: /* string */
	  switch (GTK_FUNDAMENTAL_TYPE(args[n_args].type))
	    {
	    case GTK_TYPE_STRING:
	      *((gchar**) (args[n_args].d.pointer_data)) = String_val(Field(retval, 0));
	      return;
	    }	  
	}
    }
  fprintf(stderr, "Gtk callbacks, typecheck error in return value, wanted=%d\n",
    GTK_FUNDAMENTAL_TYPE(args[n_args].type));
  failwith("Gtk.__callback_handler__");
}

void mlgtk_callback_destroy(value cbk)
{
  remove_global_root((void*)cbk);
  free((void*)cbk);
}

value mlgtk_signal_connect(value object, value name, value func)
{
  value *cbk;
  check_GtkObject_ml(object, "Gtk.Unsafe.signal_connect");
  cbk = (value*) malloc(sizeof(value));
  register_global_root(cbk);
  *cbk = func;
  return Val_int(gtk_signal_connect_interp(
    GtkObject_ml(object), String_val(name),
    (GtkCallbackMarshal) mlgtk_callback_exec, cbk,
    (GtkDestroyNotify) mlgtk_callback_destroy, 1));
}

value mlgtk_signal_disconnect(value object, value index)
{
  gtk_signal_disconnect(GtkObject_ml(object), Int_val(index));
}

/* ###### 4. Widgets */
/* ### 5.1 Alignment widget */
value mlgtk_alignment_new(value xalign, value yalign,
  value xscale, value yscale)
{
  return ml_GtkObject(
    gtk_alignment_new(Double_val(xalign), Double_val(yalign),
      Double_val(xscale), Double_val(yscale)));
}

value mlgtk_alignment_set(value alignment,
  value xalign, value yalign, value xscale, value yscale)
{
  gtk_alignment_set(GtkAlignment_ml(alignment),
    Double_val(xalign), Double_val(yalign),
    Double_val(xscale), Double_val(yscale));
  return Val_unit;
}

/* ### 5.2 Arrow widget */
value mlgtk_arrow_new(value arrow_type, value shadow_type)
{
  return ml_GtkObject(
    gtk_arrow_new(Int_val(arrow_type), Int_val(shadow_type)));
}

value mlgtk_arrow_set(value arrow, value arrow_type, value shadow_type)
{
  gtk_arrow_set(GtkArrow_ml(arrow), Int_val(arrow_type), Int_val(shadow_type));
  return Val_unit;
}

/* ### 5.3 Aspect frame widget */
value mlgtk_aspect_frame_new(value label, value xalign, value yalign,
  value ratio, value obey_child)
{
  return (value) gtk_aspect_frame_new(String_val(label), Double_val(xalign),
    Double_val(yalign), Double_val(ratio), Bool_val(obey_child));
}

value mlgtk_aspect_frame_set(value aspect_frame,
  value xalign, value yalign, value ratio, value obey_child)
{
  gtk_aspect_frame_set(GtkAspectFrame_ml(aspect_frame), Double_val(xalign),
    Double_val(yalign), Double_val(ratio), Bool_val(obey_child));
  return Val_unit;
}

/* ### 5.5 Box widget */
value mlgtk_box_pack_start(value container, value widget,
  value expand, value fill, value padding)
{
  check_GtkObject_ml(container, "Gtk.Unsafe.box_pack_start");
  check_GtkObject_ml(widget, "Gtk.Unsafe.box_pack_start");
  gtk_box_pack_start(
    GtkBox_ml(container), GtkWidget_ml(widget),
    Bool_val(expand), Bool_val(fill), Int_val(padding));
  return Val_unit;  
}

value mlgtk_box_pack_end(value container, value widget,
  value expand, value fill, value padding)
{
  check_GtkObject_ml(container, "Gtk.Unsafe.box_pack_end");
  check_GtkObject_ml(widget, "Gtk.Unsafe.box_pack_end");
  gtk_box_pack_end(
    GtkBox_ml(container), GtkWidget_ml(widget),
    Bool_val(expand), Bool_val(fill), Int_val(padding));
  return Val_unit;  
}

value mlgtk_box_set_homogeneous(value box, value homogeneous)
{
  check_GtkObject_ml(box, "Gtk.Unsafe.box_set_homogeneous");
  gtk_box_set_homogeneous(GtkBox_ml(box), Bool_val(homogeneous));
  return Val_unit;
}

value mlgtk_box_set_spacing(value box, value spacing)
{
  check_GtkObject_ml(box, "Gtk.Unsafe.box_set_spacing");
  gtk_box_set_spacing(GtkBox_ml(box), Int_val(spacing));
  return Val_unit;
}

value mlgtk_box_reorder_child(value box, value child, value pos)
{ 
  check_GtkObject_ml(box, "Gtk.Unsafe.box_reorder_child");
  check_GtkObject_ml(child, "Gtk.Unsafe.box_reorder_child");
  gtk_box_reorder_child(GtkBox_ml(box), GtkWidget_ml(child), Int_val(pos));
  return Val_unit;
}

value mlgtk_box_query_child_packing(value box, value child)
{
  gint expand, fill, padding;
  GtkPackType pack_type;
  value r;

  check_GtkObject_ml(box, "Gtk.Unsafe.box_query_child_packing");
  check_GtkObject_ml(child, "Gtk.Unsafe.box_query_child_packing");
  gtk_box_query_child_packing(GtkBox_ml(box), GtkWidget_ml(child),
    &expand, &fill, &padding, &pack_type);

  r=alloc_tuple(4);
  Field(r, 0)= Val_bool(expand);
  Field(r, 1)= Val_bool(fill);
  Field(r, 2)= Val_int(padding);
  Field(r, 3)= Val_int(pack_type);
  return r;  
}

value mlgtk_box_set_child_packing(value box, value child,
  value expand, value fill, value padding, value pack_type)
{
  check_GtkObject_ml(box, "Gtk.Unsafe.box_set_child_packing");
  check_GtkObject_ml(child, "Gtk.Unsafe.box_set_child_packing");
  gtk_box_set_child_packing(GtkBox_ml(box), GtkWidget_ml(child),
    Bool_val(expand), Bool_val(fill), Int_val(padding), Int_val(pack_type));
  return Val_unit;
}
value mlgtk_box_set_child_packing_bytecode(value *argv, int argc)
{
  return mlgtk_box_set_child_packing(argv[0], argv[1], argv[2], argv[3],
    argv[4], argv[5]);
}

/* ### 5.6 Button box widget */
value mlgtk_button_box_set_child_size_default(value width, value height)
{
  gtk_button_box_set_child_size_default(Int_val(width), Int_val(height));
  return Val_unit;
}

value mlgtk_button_box_set_child_ipadding_default(value ipad_x, value ipad_y)
{
  gtk_button_box_set_child_ipadding_default(Int_val(ipad_x), Int_val(ipad_y));
  return Val_unit;
}

value mlgtk_button_box_get_child_size_default(value dummy)
{
  int width, height;
  value r;
  gtk_button_box_get_child_size_default(&width, &height);
  r=alloc_tuple(2);
  Field(r, 0)=Val_int(width);
  Field(r, 1)=Val_int(height);
  return r;
}

value mlgtk_button_box_get_child_ipadding_default(value dummy)
{
  int ipad_x, ipad_y;
  value r;
  gtk_button_box_get_child_ipadding_default(&ipad_x, &ipad_y);
  r=alloc_tuple(2);
  Field(r, 0)=Val_int(ipad_x);
  Field(r, 1)=Val_int(ipad_y);
  return r;
}

value mlgtk_button_box_set_child_size(value widget, value width, value height)
{
  check_GtkObject_ml(widget, "Gtk.Unsafe.button_box_set_child_size");
  gtk_button_box_set_child_size(GtkButtonBox_ml(widget),
    Int_val(width), Int_val(height));
  return Val_unit;
}

value mlgtk_button_box_set_child_ipadding(value widget,
  value ipad_x, value ipad_y)
{
  check_GtkObject_ml(widget, "Gtk.Unsafe.button_box_set_child_ipadding");
  gtk_button_box_set_child_ipadding(GtkButtonBox_ml(widget),
    Int_val(ipad_x), Int_val(ipad_y));
  return Val_unit;
}

value mlgtk_button_box_set_layout(value widget, value layout_style)
{
  check_GtkObject_ml(widget, "Gtk.Unsafe.button_box_set_layout");
  gtk_button_box_set_layout(GtkButtonBox_ml(widget), Int_val(layout_style));
  return Val_unit;
}

value mlgtk_button_box_get_spacing(value widget)
{
  check_GtkObject_ml(widget, "Gtk.Unsafe.button_box_get_spacing");
  return Val_int(gtk_button_box_get_spacing(GtkButtonBox_ml(widget)));
}

value mlgtk_button_box_get_child_size(value widget)
{
  int width, height;
  value r;
  check_GtkObject_ml(widget, "Gtk.Unsafe.button_box_get_child_size");
  gtk_button_box_get_child_size(GtkButtonBox_ml(widget), &width, &height);
  r=alloc_tuple(2);
  Field(r,0)=Val_int(width);
  Field(r,1)=Val_int(height);
  return r;
}

value mlgtk_button_box_get_child_ipadding(value widget)
{
  value r;
  int ipad_x, ipad_y;
  check_GtkObject_ml(widget, "Gtk.Unsafe.button_box_get_child_ipadding");
  gtk_button_box_get_child_ipadding(GtkButtonBox_ml(widget), &ipad_x, &ipad_y);
  r=alloc_tuple(2);
  Field(r, 0)=Val_int(ipad_x);
  Field(r, 1)=Val_int(ipad_y);
  return r;
}

value mlgtk_button_box_get_layout(value widget)
{
  check_GtkObject_ml(widget, "Gtk.Unsafe.button_box_get_layout");
  return Val_int(gtk_button_box_get_layout(GtkButtonBox_ml(widget)));
}

/* ### 5.7 Button widget */
value mlgtk_button_new_with_label(value label)
{
  return ml_GtkObject(gtk_button_new_with_label(String_val(label)));
}

value mlgtk_button_released(value button)
{
  gtk_button_released(GtkButton_ml(button));
  return Val_unit;
}

value mlgtk_button_pressed(value button)
{
  gtk_button_pressed(GtkButton_ml(button));
  return Val_unit;
}

value mlgtk_button_clicked(value button)
{
  gtk_button_clicked(GtkButton_ml(button));
  return Val_unit;
}

value mlgtk_button_enter(value button)
{
  gtk_button_enter(GtkButton_ml(button));
  return Val_unit;
}

value mlgtk_button_leave(value button)
{
  gtk_button_leave(GtkButton_ml(button));
  return Val_unit;
}

/* ### 5.8 Check button widget */
value mlgtk_check_button_new_with_label(value label)
{
  return ml_GtkObject(gtk_check_button_new_with_label(String_val(label)));
}

/* ### 5.9 Check menu item widget */
value mlgtk_check_menu_item_new_with_label(value str)
{
  return ml_GtkObject(gtk_check_menu_item_new_with_label(String_val(str)));
}

value mlgtk_check_menu_item_set_state(value object,value state)
{
  check_GtkObject_ml(object, "Gtk.Unsafe.check_menu_item_set_state");
  gtk_check_menu_item_set_state(GtkCheckMenuItem_ml(object),Bool_val(state));
  return Val_unit;
}

value mlgtk_check_menu_item_toggled(value object)
{
  gtk_check_menu_item_toggled(GtkCheckMenuItem_ml(object));
  return Val_unit;
}

/* ### 5.10 Compound list */
value mlgtk_clist_new(value columns)
{
  return ml_GtkObject(gtk_clist_new(Int_val(columns)));
}

value mlgtk_clist_set_border(value clist, value shadow_type)
{
  gtk_clist_set_border(GtkCList_ml(clist), Val_int(shadow_type));
  return Val_unit;
}

value mlgtk_clist_set_selection_mode(value clist, value selection_mode)
{
  gtk_clist_set_border(GtkCList_ml(clist), Val_int(selection_mode));
  return Val_unit;
}

value mlgtk_clist_set_policy(value clist, value policy)
{
  gtk_clist_set_border(GtkCList_ml(clist), Val_int(policy));
  return Val_unit;
}

value mlgtk_clist_freeze(value clist)
{
  gtk_clist_freeze(GtkCList_ml(clist));
  return Val_unit;
}

value mlgtk_clist_thaw(value clist)
{
  gtk_clist_thaw(GtkCList_ml(clist));
  return Val_unit;
}

value mlgtk_clist_column_titles_show(value clist)
{
  gtk_clist_column_titles_show(GtkCList_ml(clist));
  return Val_unit;
}

value mlgtk_clist_column_titles_hide(value clist)
{
  gtk_clist_column_titles_hide(GtkCList_ml(clist));
  return Val_unit;
}

value mlgtk_clist_set_column_title(value clist, value column, value title)
{
  gtk_clist_set_column_title(GtkCList_ml(clist), Int_val(column), String_val(title));
  return Val_unit;
}

value mlgtk_clist_set_column_widget(value clist, value column, value widget)
{
  gtk_clist_set_column_widget(GtkCList_ml(clist), Int_val(column), GtkWidget_ml(widget));
  return Val_unit;
}

value mlgtk_clist_set_column_justification(value clist, value column, value justification)
{
  gtk_clist_set_column_justification(GtkCList_ml(clist), Int_val(column), Int_val(justification));
  return Val_unit;
}

value mlgtk_clist_set_column_width(value clist, value column, value width)
{
  gtk_clist_set_column_width(GtkCList_ml(clist), Int_val(column), Int_val(width));
  return Val_unit;
}

value mlgtk_clist_set_row_height(value clist, value height)
{
  gtk_clist_set_row_height(GtkCList_ml(clist), Int_val(height));
  return Val_unit;
}

value mlgtk_clist_moveto(value clist, value row, value column,
  value row_align, value col_align)
{
  gtk_clist_moveto(GtkCList_ml(clist), Int_val(row), Int_val(column),
    Double_val(row_align), Double_val(col_align));
  return Val_unit; 
}

value mlgtk_clist_set_text(value clist, value row, value column, value text)
{
  gtk_clist_set_text(GtkCList_ml(clist), Int_val(row), Int_val(column), String_val(text));
  return Val_unit;
}

value mlgtk_clist_set_shift(value clist, value row, value column,
  value vertical, value horizontal)
{
  gtk_clist_set_shift(GtkCList_ml(clist), Int_val(row), Int_val(column),
    Int_val(vertical), Int_val(horizontal));
  return Val_unit;
}

value mlgtk_clist_remove(value clist, value row)
{
  gtk_clist_remove(GtkCList_ml(clist), Int_val(row));
  return Val_unit;
}

value mlgtk_clist_select_row(value clist, value row, value column)
{
  gtk_clist_select_row(GtkCList_ml(clist), Int_val(row), Int_val(column));
  return Val_unit;
}

value mlgtk_clist_unselect_row(value clist, value row, value column)
{
  gtk_clist_unselect_row(GtkCList_ml(clist), Int_val(row), Int_val(column));
  return Val_unit;
}

value mlgtk_clist_clear(value clist)
{
  gtk_clist_clear(GtkCList_ml(clist));
  return Val_unit;
}

/* ### 5.11 Color selector */
value mlgtk_color_selection_set_color(value widget, value color)
{
  double color_g[4];
  check_GtkObject_ml(GtkColorSelection_ml(widget), "Gtk.Unsafe.color_selection_set_color");
  color_g[0]=Double_val(Field(color, 0));
  color_g[1]=Double_val(Field(color, 1));
  color_g[2]=Double_val(Field(color, 2));
  color_g[3]=Double_val(Field(color, 3));
  gtk_color_selection_set_color(GtkColorSelection_ml(widget), color_g);
  return Val_unit;
}

value mlgtk_color_selection_get_color(value widget)
{
  double color_g[4];
  value r;
  check_GtkObject_ml(widget, "Gtk.Unsafe.color_selection_set_color");
  color_g[3]=1.0;
  gtk_color_selection_get_color(GtkColorSelection_ml(widget), color_g);
  r=alloc_tuple(4);
  Field(r, 0)=Field(r, 1)=Field(r, 2)=Field(r, 3)=0;
  Field(r, 0)=copy_double(color_g[0]);
  Field(r, 1)=copy_double(color_g[1]);
  Field(r, 2)=copy_double(color_g[2]);
  Field(r, 3)=copy_double(color_g[3]);
  return r;
}

value mlgtk_color_selection_set_update_policy(value widget, value policy)
{
  gtk_color_selection_set_update_policy(GtkColorSelection_ml(widget),
    Int_val(policy));
  return Val_unit;
}

value mlgtk_color_selection_set_opacity(value widget, value flag)
{
  gtk_color_selection_set_opacity(GtkColorSelection_ml(widget),
    Int_val(flag));
  return Val_unit;
}

/* ### 5.12 Combo widget */
value mlgtk_combo_set_value_in_list(value combo, value val, value ok_if_empty)
{
  gtk_combo_set_value_in_list(GtkCombo_ml(combo), Int_val(val), Int_val(ok_if_empty));
  return Val_unit; 
}

value mlgtk_combo_set_use_arrows(value combo, value val)
{
  gtk_combo_set_use_arrows(GtkCombo_ml(combo), Int_val(val));
  return Val_unit;
}

value mlgtk_combo_set_use_arrows_always(value combo, value val)
{
  gtk_combo_set_use_arrows_always(GtkCombo_ml(combo), Int_val(val));
  return Val_unit;
}

value mlgtk_combo_set_case_sensitive(value combo, value val)
{
  gtk_combo_set_case_sensitive(GtkCombo_ml(combo), Int_val(val));
  return Val_unit;
}

value mlgtk_combo_set_item_string(value combo, value item, value item_value)
{
  gtk_combo_set_item_string(GtkCombo_ml(combo), GtkItem_ml(item), String_val(item_value));
  return Val_unit; 
}

value mlgtk_combo_set_popdown_strings(value combo, value strings)
{
  gtk_combo_set_popdown_strings(GtkCombo_ml(combo), (GList*) strings);
  return Val_unit;
}

/* ### 5.13 Container widget */
value mlgtk_container_border_width(value window, value width)
{
  gtk_container_border_width(GTK_CONTAINER(window), Int_val(width));
  return Val_unit;
}

value mlgtk_container_add(value container, value widget)
{
  check_GtkObject_ml(container, "Gtk.Unsafe.container_add");
  check_GtkObject_ml(widget, "Gtk.Unsafe.container_add");
  gtk_container_add(
    GtkContainer_ml(container), GtkWidget_ml(widget));
  return Val_unit;
}

value mlgtk_container_remove(value container, value widget)
{
  gtk_container_remove(GtkContainer_ml(container), GtkWidget_ml(widget));
  return Val_unit;
}

value mlgtk_container_enable_resize(value container)
{
  gtk_container_enable_resize(GtkContainer_ml(container));
  return Val_unit;
}

value mlgtk_container_disable_resize(value container)
{
  gtk_container_disable_resize(GtkContainer_ml(container));
  return Val_unit;
}

value mlgtk_container_block_resize(value container)
{
  gtk_container_block_resize(GtkContainer_ml(container));
  return Val_unit;
}

value mlgtk_container_unblock_resize(value container)
{
  gtk_container_unblock_resize(GtkContainer_ml(container));
  return Val_unit;
}

value mlgtk_container_need_resize(value container)
{
  gtk_container_need_resize(GtkContainer_ml(container));
  return Val_unit;
}

value mlgtk_container_foreach(value container, value func)
{
  value *cbk = (value*) malloc(sizeof(value));
  register_global_root(cbk);
  *cbk = func;
  gtk_container_foreach_interp(GtkContainer_ml(container),
    (GtkCallbackMarshal) mlgtk_callback_exec,
    cbk, (GtkDestroyNotify) mlgtk_callback_destroy);
  return Val_unit;
}

value mlgtk_container_focus(value container, value direction_type)
{
  gtk_container_focus(GtkContainer_ml(container), Int_val(direction_type));
  return Val_unit;
}

/* ### 5.15 Curve widget */
value mlgtk_curve_reset(value curve)
{
  gtk_curve_reset(GtkCurve_ml(curve));
  return Val_unit;
}

value mlgtk_curve_set_gamma(value curve, value gamma)
{
  gtk_curve_set_gamma(GtkCurve_ml(curve), Double_val(gamma));
  return Val_unit;
}

value mlgtk_curve_set_range(value curve, value min_x, value max_x,
  value min_y, value max_y)
{
  gtk_curve_set_range(GtkCurve_ml(curve), Double_val(min_x), Double_val(max_x),
    Double_val(min_y), Double_val(max_y));
  return Val_unit;
}

/* works only if gfloat = double */
value mlgtk_curve_get_vector(value curve, value vector)
{
  gtk_curve_get_vector(GtkCurve_ml(curve), Double_array_length(vector), (gfloat*) vector);
  return Val_unit;
}

value mlgtk_curve_set_vector(value curve, value vector)
{
  gtk_curve_set_vector(GtkCurve_ml(curve), Double_array_length(vector), (gfloat*) vector);
  return Val_unit;
}

value mlgtk_curve_set_curve_type(value curve, value type)
{
  gtk_curve_set_curve_type(GtkCurve_ml(curve), Int_val(type));
  return Val_unit;
}

/* ### 5.16 Gamma curve widget */

/* ### 5.17 Dialog widget */

/* ### 5.18 Drawing area widget */
value mlgtk_drawing_area_size(value darea,
  gint width, gint height)
{
  gtk_drawing_area_size(GtkDrawingArea_ml(darea),
    Int_val(width), Int_val(height));
  return Val_unit;
}

/* ### 5.19 Entry widget */
value mlgtk_entry_new(value dummy)
{
  return ml_GtkObject(gtk_entry_new());
}

value mlgtk_entry_new_with_max_length(value len)
{
  return ml_GtkObject( gtk_entry_new_with_max_length((guint16) Int_val(len)) ) ;
}

value mlgtk_entry_set_text(value entry,value text)
{
  check_GtkObject_ml(entry, "Gtk.Unsafe.entry_set_text");
  gtk_entry_set_text(GtkEntry_ml(entry),String_val(text ));
  return Val_unit;
}

value mlgtk_entry_append_text(value entry, value text)
{ 
  gtk_entry_append_text(GtkEntry_ml(entry),String_val(text ));
  return Val_unit;
}

value mlgtk_entry_prepend_text(value entry, value text)
{ 
  gtk_entry_prepend_text(GtkEntry_ml(entry),String_val(text ));
  return Val_unit;
}

value mlgtk_entry_set_position(value entry, value position)
{
  gtk_entry_set_position(GtkEntry_ml(entry), Int_val(position));
  return Val_unit;
}

value mlgtk_entry_set_visibility(value entry, value visibility)
{
  gtk_entry_set_visibility(GtkEntry_ml(entry), Bool_val(visibility));
  return Val_unit;
}

value mlgtk_entry_get_text(value entry)
{
  check_GtkObject_ml(entry, "Gtk.Unsafe.entry_get_text");
  return copy_string( gtk_entry_get_text(GtkEntry_ml(entry)) );
}

/* ### 5.23 Frame widget */
value mlgtk_frame_new(value label)
{
  return ml_GtkObject(gtk_frame_new(String_val(label)));
}

value mlgtk_frame_set_shadow_type(value frame,value type)
{
  check_GtkObject_ml(frame, "Gtk.Unsafe.frame_set_shadow_type");
  check_GtkObject_ml(type, "Gtk.Unsafe.frame_set_shadow_type");
  gtk_frame_set_shadow_type(GtkFrame_ml(frame),Int_val(type));
}

/* ### 5.25 Horizontal box widget */
value mlgtk_hbox_new(value homogeneous, value spacing)
{
  return ml_GtkObject(gtk_hbox_new(Int_val(homogeneous), Int_val(spacing)));
}

/* ### 5.27 Horizontal paned widget */
value mlgtk_hpaned_new(value dummy)
{
  return ml_GtkObject(gtk_hpaned_new());
}

/* ### 5.30 Horizontal scrollbar widget */

/* ### 5.31 Horizontal separator widget */
value mlgtk_hseparator_new(value dummy)
{
  return ml_GtkObject(gtk_hseparator_new ());
}

/* ### 5.32 Image widget */

/* ### 5.33 Input dialog widget */

/* ### 5.34 Item widget */

/* ### 5.35 Label widget */
value mlgtk_label_new(value str)
{
  return ml_GtkObject(gtk_label_new(String_val(str)));
}

/* ### 5.36 List widget */
value mlgtk_list_new(value dummy)
{
  return ml_GtkObject(gtk_list_new());
}

value mlgtk_list_insert_items(value list, value items, value index)
{
  gtk_list_insert_items(GtkList_ml(list), (GList*) items, Int_val(index));
  return Val_unit;
}


value mlgtk_list_append_items(value list, value items)
{
  gtk_list_append_items(GtkList_ml(list), (GList*) items);
  return Val_unit;
}

/* ### 5.37 List item widget */
value mlgtk_list_item_new_with_label(value s)
{
  return ml_GtkObject(gtk_list_item_new_with_label(String_val(s)));
}

/* ### 5.38 Menu widget */
value mlgtk_menu_new(value dummy)
{
  return ml_GtkObject(gtk_menu_new());
}

value mlgtk_menu_append(value obj1, value obj2)
{
  gtk_menu_append(GtkMenu_ml(obj1), GtkWidget_ml(obj2));
  return Val_unit;
}

/* ### 5.39 Menu bar widget */
value mlgtk_menu_bar_new(value dummy)
{
  return ml_GtkObject(gtk_menu_bar_new());
}

value mlgtk_menu_bar_append(value obj1, value obj2)
{
  gtk_menu_bar_append(GtkMenuBar_ml(obj1), GtkWidget_ml(obj2));
  return Val_unit;
}

/* ### 5.40 Menu item widget */
value mlgtk_menu_item_new_with_label(value str)
{
  return ml_GtkObject(gtk_menu_item_new_with_label(String_val(str)));
}

value mlgtk_menu_item_set_submenu(value obj1, value obj2)
{
  gtk_menu_item_set_submenu(GtkMenuItem_ml(obj1), GtkWidget_ml(obj2));
  return Val_unit;
}

value mlgtk_menu_item_select (value menu_item)
{
  check_GtkObject_ml(menu_item, "Gtk.Unsafe.item_select");
  gtk_menu_item_select(GtkMenuItem_ml(menu_item));
  return Val_unit;
}  

value mlgtk_menu_item_deselect (value menu_item)
{
  check_GtkObject_ml(menu_item, "Gtk.Unsafe.item_deselect");
  gtk_menu_item_deselect(GtkMenuItem_ml(menu_item));
  return Val_unit;
}  

/* ### 5.41 Menu shell widget */

/* ### 5.42 Misc widget */

/* ### 5.43 Notebook widget */
value mlgtk_notebook_new(value dummy)
{
  return ml_GtkObject(gtk_notebook_new());
}

value mlgtk_notebook_append_page(value notebook,value page,value label)
{
  check_GtkObject_ml(notebook, "Gtk.Unsafe.notebook_append_page");
  check_GtkObject_ml(page, "Gtk.Unsafe.notebook_append_page");
  check_GtkObject_ml(label, "Gtk.Unsafe.notebook_append_page");
  gtk_notebook_append_page(GtkNotebook_ml(notebook),GtkWidget_ml(page),
			   GtkWidget_ml(label));
  return Val_unit;
}

value mlgtk_notebook_prepend_page(value notebook,value page,value label)
{
  check_GtkObject_ml(notebook, "Gtk.Unsafe.notebook_prepend_page");
  check_GtkObject_ml(page, "Gtk.Unsafe.notebook_prepend_page");
  check_GtkObject_ml(label, "Gtk.Unsafe.notebook_prepend_page");
  gtk_notebook_prepend_page(GtkNotebook_ml(notebook),GtkWidget_ml(page),
			   GtkWidget_ml(label));
  return Val_unit;
}


value mlgtk_notebook_insert_page(value notebook,value page,value label,
				 value pos)
{
  check_GtkObject_ml(notebook, "Gtk.Unsafe.notebook_insert_page");
  check_GtkObject_ml(page, "Gtk.Unsafe.notebook_insert_page");
  check_GtkObject_ml(label, "Gtk.Unsafe.notebook_insert_page");
  gtk_notebook_insert_page(GtkNotebook_ml(notebook),GtkWidget_ml(page),
			   GtkWidget_ml(label),Int_val(pos));
  return Val_unit;
}

value mlgtk_notebook_remove_page(value notebook,value page_num)
{
  check_GtkObject_ml(notebook, "Gtk.Unsafe.notebook_remove_page");
  gtk_notebook_remove_page(GtkNotebook_ml(notebook), Int_val(page_num));
  return Val_unit;
}

value mlgtk_notebook_set_page(value notebook,value page_num)
{
  check_GtkObject_ml(notebook, "Gtk.Unsafe.notebook_set_page");
  gtk_notebook_set_page(GtkNotebook_ml(notebook), Int_val(page_num));
  return Val_unit;
}

value mlgtk_notebook_next_page(value notebook)
{
  check_GtkObject_ml(notebook, "Gtk.Unsafe.notebook_next_page");
  gtk_notebook_next_page(GtkNotebook_ml(notebook));
  return Val_unit;
}

value mlgtk_notebook_prev_page(value notebook)
{
  check_GtkObject_ml(notebook, "Gtk.Unsafe.notebook_prev_page");
  gtk_notebook_prev_page(GtkNotebook_ml(notebook));
  return Val_unit;
}

value mlgtk_notebook_set_tab_pos(value notebook,value pos)
{
  check_GtkObject_ml(notebook, "Gtk.Unsafe.notebook_set_tab_pos");
  gtk_notebook_set_tab_pos(GtkNotebook_ml(notebook), Int_val(pos));
  return Val_unit;
}

value mlgtk_notebook_set_show_tabs(value notebook,value show_tabs)
{
  check_GtkObject_ml(notebook, "Gtk.Unsafe.notebook_set_show_tabs");
  gtk_notebook_set_show_tabs(GtkNotebook_ml(notebook), Int_val(show_tabs));
  return Val_unit;
}

value mlgtk_notebook_set_show_border(value notebook,value show_border)
{
  check_GtkObject_ml(notebook, "Gtk.Unsafe.notebook_set_show_border");
  gtk_notebook_set_show_border(GtkNotebook_ml(notebook), Int_val(show_border));
  return Val_unit;
}

/* ### 5.45 Paned widget */
value mlgtk_paned_add1(value ml_paned,value ml_widget)
{
  check_GtkObject_ml(ml_paned, "Gtk.Unsafe.paned_add1");
  check_GtkObject_ml(ml_widget, "Gtk.Unsafe.paned_add1");
  gtk_paned_add1(GtkPaned_ml(ml_paned),GtkWidget_ml(ml_widget));
  return Val_unit;
}


value mlgtk_paned_add2(value ml_paned,value ml_widget)
{
  check_GtkObject_ml(ml_paned, "Gtk.Unsafe.paned_add2");
  check_GtkObject_ml(ml_widget, "Gtk.Unsafe.paned_add2");
  gtk_paned_add2(GtkPaned_ml(ml_paned),GtkWidget_ml(ml_widget));
  return Val_unit;
}


value mlgtk_paned_handle_size(value ml_paned,value size)
{
  check_GtkObject_ml(ml_paned, "Gtk.Unsafe.paned_handle_size");
  gtk_paned_handle_size(GtkPaned_ml(ml_paned),(guint16)Int_val(size));
  return Val_unit;
}

value mlgtk_paned_gutter_size(value ml_paned,value size)
{
  check_GtkObject_ml(ml_paned, "Gtk.Unsafe.paned_gutter_size");
  gtk_paned_gutter_size(GtkPaned_ml(ml_paned),(guint16)Int_val(size));
  return Val_unit;
}

/* ### 5.46 Pixmap widget */

/* ### 5.47 Preview widget */

/* ### 5.48 Progress bar widget */
value mlgtk_progress_bar_new(value dummy)
{
  return ml_GtkObject(gtk_progress_bar_new ());
}

value mlgtk_progress_bar_update(value pbar, value ratio)
{
  gtk_progress_bar_update(GtkProgressBar_ml(pbar), (gfloat) Double_val(ratio));
  return Val_unit;
}

/* ### 5.49 Radio button widget */
value mlgtk_radio_button_new_with_label(value group, value label)
{
  return ml_GtkObject(gtk_radio_button_new_with_label(
    optionptr_ml(group), String_val(label)));
}

/* ### 5.50 Radio menu item widget */
value mlgtk_radio_menu_item_new_with_label(value group, value label)
{
  return ml_GtkObject(gtk_radio_menu_item_new_with_label(
    optionptr_ml(group), String_val(label)));
}

/* ### 5.55 Scrolled window widget */
value mlgtk_scrolled_window_new (value dummy)
{
  return ml_GtkObject(gtk_scrolled_window_new (NULL, NULL));
  /* FIXME : should take 2 arguments GtkAdjustment */
}

/* ### 5.59 Text widget */
value mlgtk_text_new(value dummy)
{
  GtkWidget* text;
  text = gtk_text_new (NULL,NULL); /* pcuoq : NULL at least until GtkAdjustment 
				      is documented */
  return ml_GtkObject(text);
}

value mlgtk_text_set_editable(value ml_text,value flag)
{
  GtkWidget* text;
  check_GtkObject_ml(ml_text, "Gtk.Unsafe.text_set_editable");
  gtk_text_set_editable (GtkText_ml(ml_text), Bool_val(flag));
  return Val_unit;
}

/* Function: void gtk_text_insert (GtkText *TEXT, GdkFont *FONT,
           GdkColor *FORE, GdkColor *BACK, char *CHARS, gint LENGTH) */
value mlgtk_text_insert(value ml_text, value ml_font, value ml_chars, value start, value length)
{
  check_GtkObject_ml(ml_text, "Gtk.Unsafe.text_insert");
  fprintf(stderr,"debug : %d %d\n",Int_val(start), Int_val(length));
  gtk_text_insert(GtkText_ml(ml_text),optionptr_ml(ml_font),NULL,NULL,
		  String_val(ml_chars)+Int_val(start), Int_val(length) );
  return Val_unit;
}

value mlgtk_text_freeze(value mltext)
{
  check_GtkObject_ml(mltext, "Gtk.Unsafe.text_freeze");
  gtk_text_freeze(GtkText_ml(mltext));
  return Val_unit;
}

value mlgtk_text_thaw(value mltext)
{
  check_GtkObject_ml(mltext, "Gtk.Unsafe.text_thaw");
  gtk_text_thaw(GtkText_ml(mltext));
  return Val_unit;
}

value mlgtk_text_forward_delete(value mltext,value n)
{
  check_GtkObject_ml(mltext, "Gtk.Unsafe.text_forward_delete");
  return Val_int( gtk_text_forward_delete(GtkText_ml(mltext),Int_val(n)) ) ;
}

value mlgtk_text_backward_delete(value mltext,value n)
{
  check_GtkObject_ml(mltext, "Gtk.Unsafe.text_backward_delete");
  return Val_int( gtk_text_backward_delete(GtkText_ml(mltext),Int_val(n)) ) ;
}

/* ### 5.60 Toggle button widget */
value mlgtk_toggle_button_new_with_label(value label)
{
  return ml_GtkObject(gtk_toggle_button_new_with_label(String_val(label)));
}

value mlgtk_toggle_button_set_state(value button, value state)
{
  check_GtkObject_ml(widget, "Gtk.Unsafe.toggle_button_set_state");  
  gtk_toggle_button_set_state(GtkToggleButton_ml(button), Int_val(state));
  return Val_unit;
}

/* ### 5.62 Tool tips widget */
value mlgtk_tooltips_new(value dummy)
{
  return ml_GtkObject(gtk_tooltips_new ());
}

value mlgtk_tooltips_set_tip (value tooltips, value widget, value str,value private)
{
  check_GtkObject_ml(tooltips, "Gtk.Unsafe.tooltips_set_tip");
  check_GtkObject_ml(widget, "Gtk.Unsafe.tooltips_set_tip");
  gtk_tooltips_set_tip (GtkTooltips_ml(tooltips),
    GtkWidget_ml(widget), String_val(str), String_val(private));
  return Val_unit;
}

/* ### 5.65 Vertical box widget */
value mlgtk_vbox_new(value homogeneous, value spacing)
{
  return ml_GtkObject(gtk_vbox_new(Int_val(homogeneous), Int_val(spacing)));
}

/* ### 5.67 Vertical paned widget */
value mlgtk_vpaned_new(value dummy)
{
  return ml_GtkObject(gtk_vpaned_new());
}

/* ### 5.72 Vertical separator widget */
value mlgtk_vseparator_new(value dummy)
{
  return ml_GtkObject(gtk_vseparator_new ());
}

/* ### 5.73 Base widget */
value mlgtk_widget_show(value widget)
{
  check_GtkObject_ml(widget, "Gtk.Unsafe.widget_show");
  gtk_widget_show(GtkWidget_ml(widget));
  return Val_unit;
}

value mlgtk_widget_hide(value widget)
{
  check_GtkObject_ml(widget, "Gtk.Unsafe.widget_hide");
  gtk_widget_hide(GtkWidget_ml(widget));
  return Val_unit;
}

value mlgtk_widget_realize(value mlwidget)
{
  check_GtkObject_ml(mlwidget, "Gtk.Unsafe.widget_realize");
  gtk_widget_realize(GtkWidget_ml(mlwidget));
  return Val_unit;
}

value mlgtk_widget_unrealize(value mlwidget)
{
  check_GtkObject_ml(mlwidget, "Gtk.Unsafe.widget_unrealize");
  gtk_widget_unrealize(GtkWidget_ml(mlwidget));
  return Val_unit;
}

value mlgtk_widget_set_events(value widget, value events)
{
  gtk_widget_set_events(GtkWidget_ml(widget), mlgtk_convert_events(events));
  return Val_unit;
}


value mlgtk_widget_set_usize (value widget, value x, value y)
{
  check_GtkObject_ml(widget, "Gtk.Unsafe.widget_set_usize");
  gtk_widget_set_usize(GtkWidget_ml(widget),Int_val(x),Int_val(y));
}

/* ### 5.75 Window widget */
value mlgtk_window_new(value window_type)
{
  return ml_GtkObject(gtk_window_new(Int_val(window_type)));
}

value mlgtk_window_set_title(value window, value title)
{
  check_GtkObject_ml(window, "Gtk.Unsafe.window_set_title");
  gtk_window_set_title(GtkWindow_ml(window), String_val(title));
  return Val_unit;
}

/* ### Editable widget */
value 
mlgtk_editable_insert_text(value editable,value text,value length,value position)
{
  gint gposition;
  check_GtkObject_ml(editable, "Gtk.Unsafe.editable_insert_text");
  gposition=Int_val(position);
  gtk_editable_insert_text( GtkEditable_ml(editable), String_val(text),
			     Int_val(length), &gposition );
  return Val_int(gposition);
}

value mlgtk_editable_select_region(value editable,value start,value end)
{
  check_GtkObject_ml(editable, "Gtk.Unsafe.editable_select_region"); 
  gtk_editable_select_region( GtkEditable_ml(editable), Int_val(start), Int_val(end) );
  return Val_unit;
}

value mlgtk_editable_get_chars(value editable,value start,value end)
{
  check_GtkObject_ml(editable, "Gtk.Unsafe.editable_get_chars"); 
  return copy_string( gtk_editable_get_chars( GtkEditable_ml(editable), Int_val(start), Int_val(end) ) );
}

value mlgtk_object_destroy(value object)
{
  struct callback *cbk, *nxt;
  check_GtkObject_ml(object, "Gtk.Unsafe.object_destroy");
  gtk_signal_handlers_destroy(GtkObject_ml(object));
  gtk_object_destroy(GtkObject_ml(object));
  /*GtkObject_ml(object) = NULL;*/
  return Val_unit;
}

/************************
ACTIONS
************************/
value mlgtk_timeout_add(value delay, value func)
{
  value *cbk;
  cbk = (value*) malloc(sizeof(value));
  register_global_root(cbk);
  *cbk = func;
  return Val_int(gtk_timeout_add_interp(Int_val(delay),
    (GtkCallbackMarshal) mlgtk_callback_exec, cbk,
    (GtkDestroyNotify) mlgtk_callback_destroy));
}

value mlgtk_timeout_remove(value index)
{
  gtk_timeout_remove(Int_val(index));
  return Val_unit;
}

value mlgtk_main_quit(value dummy)
{
  gtk_main_quit();
  return Val_unit;
}

int main(int argc, char **argv)
{
 gtk_set_locale ();
 
  gtk_init(&argc, &argv);
  caml_main(argv);
  /*mlgdk_init();*/
  gtk_main();
  return 0;
}


gint mlgtk_convert_events(value events)
{
  gint r=0;
  while (events!= 1)
    {
      r |= 2<< (Int_val(Field(events,0)));
      events = Field(events, 1);
    }
  return r;
}

